#!/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path

from slpkg.config import config_load
from slpkg.repositories import Repositories
from slpkg.utilities import Utilities

logger = logging.getLogger(__name__)


class Changelog:  # pylint: disable=[R0902]

    """Manages and prints changelog information for specified repositories.

    This class handles the retrieval and display of changelog entries.
    It allows filtering changelog content based on a user-provided query
    and includes terminal-based pagination for large outputs.
    """

    def __init__(self, options: dict[str, bool], repository: str, query: str) -> None:
        logger.debug("Initializing Changelog module with options: %s, repository: %s, query: '%s'",
                     options, repository, query)
        self.options = options
        self.repository = repository
        self.query = query

        self.repos = Repositories()
        self.utils = Utilities()
        self.columns, self.rows = shutil.get_terminal_size()
        logger.debug("Terminal size: %sx%s", self.columns, self.rows)

        self.repo_path: Path = self.repos.repositories[self.repository]['path']
        self.changelog_txt: str = self.repos.repositories[self.repository]['changelog_txt']
        self.changelog_file: Path = Path(self.repo_path, self.changelog_txt)
        logger.debug("Changelog file path: %s", self.changelog_file)

        self.option_for_edit: bool = options.get('option_edit', False)
        logger.debug("Option for edit enabled: %s", self.option_for_edit)

    def run(self) -> None:
        """Run changelog methods"""
        logger.info("Changelog run method called. Option for edit: %s", self.option_for_edit)
        if self.option_for_edit:
            self.edit_changelog()
        else:
            self.display_changelog()

    def display_changelog(self) -> None:
        """Prints repository changelog."""
        logger.info("Displaying changelog for repository '%s' with query: '%s'.", self.repository, self.query)
        days: tuple[str, str, str, str, str, str, str] = ('Mon ', 'Tue ', 'Wed ', 'Thu ', 'Fri ', 'Sat ', 'Sun ')
        # Console colors.
        back_white: str = config_load.back_white
        black: str = config_load.black
        green: str = config_load.green
        endc: str = config_load.endc

        # Initialize a counter for lines actually printed to the terminal.
        printed_lines_count: int = 0
        logger.debug("Initialized printed_lines_count to 0.")

        try:
            # Read changelog file line by line.
            lines = self.utils.read_text_file(self.changelog_file)
            logger.debug("Read %d lines from changelog file: %s", len(lines), self.changelog_file)

            for line_from_file in lines:
                if self.query.lower() in line_from_file.lower():  # Case-insensitive query match.
                    display_line = line_from_file  # Use a mutable variable for display.

                    if any(day in display_line for day in days):
                        display_line = f'{green}{display_line}{endc}'  # Apply color for date lines.
                        logger.debug("Applied green color to date line: '%s'", line_from_file.strip())

                    print(display_line, end='')  # Print to console, keep newline from file.
                    printed_lines_count += 1  # Increment only for lines actually printed.
                    logger.debug("Printed line. Current printed_lines_count: %d", printed_lines_count)

                    # Implement pagination
                    # Prompt if we've printed a full page AND it's not the very first line (to avoid prompt on 1st line if rows-1 is 1)
                    if printed_lines_count > 0 and (printed_lines_count % (self.rows - 1)) == 0:
                        user_input: str = input(f'{back_white}{black}[-- More --] Press Enter to continue, q to quit:{endc}')
                        logger.debug("Pagination prompt displayed. User input: '%s'", user_input)
                        if user_input.lower() in ['q', 'quit', ]:
                            print('Operation aborted by the user.')
                            logger.info("Changelog display aborted by user.")
                            sys.exit(0)
            logger.info("Finished displaying changelog.")
        except FileNotFoundError:
            logger.error("Changelog file not found: %s", self.changelog_file)
            print(f"Error: Changelog file not found at '{self.changelog_file}'")
            sys.exit(1)
        except Exception as e:  # pylint: disable=[W0718]
            logger.critical("An unexpected error occurred while displaying changelog: %s", e, exc_info=True)
            print(f"Error: An unexpected error occurred: {e}")
            sys.exit(1)

    def edit_changelog(self) -> None:
        """Edit changelog file with system editor."""
        editor = os.environ.get('EDITOR', '')
        if not editor:
            logger.error("EDITOR environment variable is not set. Cannot open external editor for changelog.")
            print("Error: EDITOR environment variable is not set. Cannot open editor.")
            sys.exit(1)

        command = [editor, str(self.changelog_file)]
        logger.info("Opening changelog file '%s' with external editor: '%s'", self.changelog_file, editor)
        try:
            subprocess.run(command, check=True)  # check=True will raise CalledProcessError on non-zero exit codes.
            logger.info("External editor exited successfully after editing changelog.")
        except subprocess.CalledProcessError as e:
            logger.error("External editor command failed with exit code %s: %s", e.returncode, e, exc_info=True)
            print(f"Error: External editor failed with exit code {e.returncode}")
            sys.exit(e.returncode)
        except FileNotFoundError:
            logger.error("External editor '%s' not found. Please ensure it is installed and in your PATH.", editor, exc_info=True)
            print(f"Error: External editor '{editor}' not found. Please ensure it is installed and in your PATH.")
            sys.exit(1)
        except Exception as e:  # pylint: disable=[W0718]
            logger.critical("An unexpected error occurred while opening external editor for changelog: %s", e, exc_info=True)
            print(f"Error: An unexpected error occurred with editor: {e}")
            sys.exit(1)

        sys.exit(0)
